﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Audio;
using UnityEngine.SceneManagement;


// -----------------------------------------------------------------------------------------
// CLASS	:	MixerGroupInfo
// DESC		:	Wraps an AudioMixerGroup in Unity's AudioMixer. Contains the name of the
//				group (which is also its exposed volume paramater), the group itself
//				and an IEnumerator for doing track fades over time.
// -----------------------------------------------------------------------------------------
public class MixerGroupInfo
{
    public string Name = string.Empty;
    public AudioMixerGroup Group = null;
    public IEnumerator TrackFader = null;
}

// ------------------------------------------------------------------------------------------
// CLASS	:	AudioPoolItem
// DESC		:	Describes an audio entity in our pooling system.
// ------------------------------------------------------------------------------------------
public class AudioPoolItem
{
    public GameObject GameObject = null;
    public Transform Transform = null;
    public AudioSource AudioSource = null;
    public float Unimportance = float.MaxValue;
    public bool Playing = false;
    public IEnumerator Coroutine = null;
    public ulong ID = 0;
}




// ----------------------------------------------------------------------------------------
// CLASS	:	AudioManager
// DESC		: 	Provides pooled one-shot functionality with priority system and also
//				wraps the Unity Audio Mixer to make easier manipulation of audiogroup volumes. 
// ---------------------------------------------------------------------------------------- 
public class AudioManager : MonoBehaviour
{
    // Statics
    private static AudioManager _instance = null;
    public static AudioManager Get
    {
        get
        {
            if (_instance == null)
                _instance = (AudioManager)FindObjectOfType(typeof(AudioManager));
            return _instance;
        }
    }


    // Inspector Assigned Variables
    [SerializeField]
    AudioMixer _mixer = null;
    [SerializeField]
    int _maxSounds = 32;


    // Private Variables
    Dictionary<string, MixerGroupInfo> _groups = new Dictionary<string, MixerGroupInfo>();
    List<AudioPoolItem> _pool = new List<AudioPoolItem>();
    Dictionary<ulong, AudioPoolItem> _activePool = new Dictionary<ulong, AudioPoolItem>();
    ulong _idGiver = 0;
    Transform _listenerPos = null;


    // AudioPool "parent" locations  // TO DO:  Change to PRIVATE once I can see they work, if they can be gotten in script instead!
    public Transform MenuAudioPool;
    public Transform InGameAudioPool;



    void Awake()
    {
        // We reference the AudioMixer through the AudioManager in the BaseGame prefab, 
        // but if we wanted we could load it from our Resources folder if need be, like so...
        //_mixer = Resources.Load<AudioMixer>("AudioMixer/MasterMixer");

        // Return if we have no valid mixer reference
        if (!_mixer) return;

        // Fetch all the groups in the mixer - These will be our mixer tracks
        AudioMixerGroup[] groups = _mixer.FindMatchingGroups(string.Empty);

        // Create our mixer tracks based on group name (a "Track" is what we will sometimes be calling a "MixerGroup")
        foreach (AudioMixerGroup group in groups)
        {
            MixerGroupInfo groupInfo = new MixerGroupInfo();
            groupInfo.Name = group.name;
            groupInfo.Group = group;
            groupInfo.TrackFader = null;
            // add this groupInfo to the _groups dictionary
            _groups[group.name] = groupInfo;
        }



        // Generate AudioPool (and place them all in the MenuAudioPool for starters)
        // Note: Make sure MenuAudioPool exists, and get reference to it (hack: it is just pre-created in the hierarchy currently)
        
        for (int i = 0; i < _maxSounds; i++)
        {
            // Create GameObject and assigned AudioSource and Parent
            GameObject go = new GameObject("Pool Item");
            AudioSource audioSource = go.AddComponent<AudioSource>();
            go.transform.parent = MenuAudioPool;

            // Create and configure Pool Item
            AudioPoolItem poolItem = new AudioPoolItem();
            poolItem.GameObject = go;
            poolItem.AudioSource = audioSource;
            poolItem.Transform = go.transform;
            poolItem.Playing = false;
            go.SetActive(false);
            _pool.Add(poolItem);

        }
    }


    // ------------------------------------------------------------------------------
    // Name	:	OnEnable
    // Desd	:	Register OnSceneLoaded Event
    // ------------------------------------------------------------------------------
    void OnEnable()
    {
        SceneManager.sceneLoaded += OnSceneLoaded;
    }


    // ------------------------------------------------------------------------------
    // Name	:	OnDisable
    // Desd	:	UnRegister OnSceneLoaded Event
    // ------------------------------------------------------------------------------
    void OnDisable()
    {
        SceneManager.sceneLoaded -= OnSceneLoaded;
    }


    // ------------------------------------------------------------------------------
    // Name	:	OnSceneLoaded
    // Desc	:	A new scene has just been loaded so we need to find the listener
    // ------------------------------------------------------------------------------
    void OnSceneLoaded(Scene scene, LoadSceneMode mode)
    {
        AudioListener listener = FindObjectOfType<AudioListener>();
        if (listener)
            _listenerPos = listener.transform;

        // get reference to InGameAudioPool parent
        if (GameManager.Get.gameState == GameManager.GameState.InGame)
        {
            FindGameAudioPoolParent();
        }
        
    }



    public void FindGameAudioPoolParent()
    {
        if (GameManager.Get.gameState == GameManager.GameState.InGame)
        {
            // set reference to the InGameAudioPool     // TO DO: Check if it exists, and make this object if it doesn't exist  (???)
            InGameAudioPool = GameObject.Find("InGameAudioPool").transform;
        }
    }



    // ------------------------------------------------------------------------------
    // Name	:	GetTrackVolume
    // Desc	:	Returns the volume of the AudioMixerGroup assign to the passed track.
    //			AudioMixerGroup MUST expose its volume variable to script for this to
    //			work and the variable MUST be the same as the name of the group
    // ------------------------------------------------------------------------------
    public float GetTrackVolume(string mixerGroupName)
    {
        MixerGroupInfo groupInfo;
        if (_groups.TryGetValue(mixerGroupName, out groupInfo))
        {
            float volume;
            _mixer.GetFloat(mixerGroupName, out volume);
            return volume;
        }

        return float.MinValue;
    }


    public AudioMixerGroup GetMixerGroupFromMixerGroupName(string mixerGroupName)
    {
        MixerGroupInfo groupInfo;
        if (_groups.TryGetValue(mixerGroupName, out groupInfo))
        {
            return groupInfo.Group;
        }

        return null;
    }



    // ------------------------------------------------------------------------------
    // Name	:	SetTrackVolume
    // Desc	:	Sets the volume of the AudioMixerGroup assigned to the passed track.
    //			AudioMixerGroup MUST expose its volume variable to script for this to
    //			work and the variable MUST be the same as the name of the group.
    //			If a fade time is given a coroutine will be used to perform the fade.
    // ------------------------------------------------------------------------------
    public void SetTrackVolume(string mixerGroupName, float volume, float fadeTime = 0.0f)
    {
        if (!_mixer) return;
        MixerGroupInfo groupInfo;
        if (_groups.TryGetValue(mixerGroupName, out groupInfo))
        {
            // Stop any coroutine that might be in the middle of fading this track
            if (groupInfo.TrackFader != null) StopCoroutine(groupInfo.TrackFader);

            // if we don't want a fade time at all, just set MixerGroup Volume directly & immediately
            if (fadeTime <= 0.0f)
                _mixer.SetFloat(mixerGroupName, volume);
            // else, set fader over time with coroutine
            else
            {
                groupInfo.TrackFader = SetTrackVolumeInternal(mixerGroupName, volume, fadeTime);
                StartCoroutine(groupInfo.TrackFader);
            }
        }

    }



    // -----------------------------------------------------------------------------------------
    // Name	:	SetTrackVolumeInternal - COROUTINE
    // Desc	:	Used by SetTrackVolume to implement a fade between volumes of a track over time.
    // -----------------------------------------------------------------------------------------
    protected IEnumerator SetTrackVolumeInternal(string mixerGroupName, float volume, float fadeTime)
    {
        float startVolume = 0.0f;
        float timer = 0.0f;
        _mixer.GetFloat(mixerGroupName, out startVolume);

        while (timer < fadeTime)
        {
            timer += Time.unscaledDeltaTime;
            // TO DO:  MUZ:  May want to use another time system, or pass in the option into this function to use Unscaled or Scaled time, 
            // depending on what needs to be done in-game.
            // Note that, if you wanted to fade the volume mixer faders over time, in the game scene, and your game had a slow-mo feature, 
            // then you'd want your faders to also move in slow-mo in that particular example.  Most of the time, you would want to use UnscaledTime, though.
            _mixer.SetFloat(mixerGroupName, Mathf.Lerp(startVolume, volume, timer / fadeTime));
            yield return null;
        }

        _mixer.SetFloat(mixerGroupName, volume);
    }



    // -------------------------------------------------------------------------------
    // Name	:	ConfigurePoolObject
    // Desc	:	Used internally to CONFIGURE & PLAY an audio source pool object
    // -------------------------------------------------------------------------------
    protected ulong ConfigurePoolObject(int poolIndex, string mixerGroupName, AudioClip clip, Vector3 position, float volume, float pitch, float spatialBlend, float unimportance)
    {
        // If poolIndex is out of range abort request
        if (poolIndex < 0 || poolIndex >= _pool.Count) return 0;

        // Get the pool item
        AudioPoolItem poolItem = _pool[poolIndex];

        // Generate new ID so we can stop it later if we want to
        _idGiver++;

        // Configure the audio source's position and volume
        AudioSource source = poolItem.AudioSource;
        source.clip = clip;
        source.volume = volume;
        source.pitch = pitch;

        source.spatialBlend = spatialBlend;
        // hack so that doppler effect is never used
        source.dopplerLevel = 0f;
        // TO DO:  Would be good to pass in a custom MaxDistance for each sound, or maybe just hack each one down to 100m or so, instead of a very large 500 default value
        //source.maxDistance = 50f;


        // Assign to requested audio group/track
        source.outputAudioMixerGroup = _groups[mixerGroupName].Group;


        // Set the parent relative to the context it is being played in (so in-game sounds can be paused when going to pause menu, for example)
        if (GameManager.Get.gameState == GameManager.GameState.MainMenu || GameManager.Get.gameState == GameManager.GameState.PauseMenu)
        {
            poolItem.GameObject.transform.parent = MenuAudioPool;
        }
        else if (GameManager.Get.gameState == GameManager.GameState.InGame)
        {
            poolItem.GameObject.transform.parent = InGameAudioPool;
        }


        // Position source at requested position
        source.transform.position = position;

        // Enable GameObject and record that it is now playing
        poolItem.Playing = true;
        poolItem.Unimportance = unimportance;
        poolItem.ID = _idGiver;
        poolItem.GameObject.SetActive(true);
        source.Play();
        poolItem.Coroutine = StopSoundDelayed(_idGiver, source.clip.length);
        StartCoroutine(poolItem.Coroutine);

        // Add this sound to our active pool Dictionary with its unique id
        _activePool[_idGiver] = poolItem;

        // Return the id to the caller
        return _idGiver;
    }



    // -------------------------------------------------------------------------------
    // Name	:	StopSoundDelayed
    // Desc	:   Stop a one shot sound from playing after a number of seconds
    // -------------------------------------------------------------------------------
    protected IEnumerator StopSoundDelayed(ulong id, float duration)
    {
        yield return new WaitForSeconds(duration);
        AudioPoolItem activeSound;

        // If this if exists in our active pool
        if (_activePool.TryGetValue(id, out activeSound))
        {
            activeSound.AudioSource.Stop();
            activeSound.AudioSource.clip = null;
            activeSound.GameObject.SetActive(false);
            _activePool.Remove(id);

            // Make it available again
            activeSound.Playing = false;
        }
    }



    // -------------------------------------------------------------------------------
    // Name	:	PlayOneShotSound
    // Desc	:	Scores the priority of the sound and search for an unused pool item to use as the audio source. 
    //			If one is not available an audio source with a lower priority will be killed and reused.
    // -------------------------------------------------------------------------------
    public ulong PlayOneShotSound(string mixerGroupName, AudioClip clip, Vector3 position, float volume, float pitch, float spatialBlend, int priority = 128)
    {
        // Do nothing if mixer group does not exist, clip is null or volume is zero
        if (!_groups.ContainsKey(mixerGroupName) || clip == null || volume.Equals(0.0f)) return 0;

        float unimportance = (_listenerPos.position - position).sqrMagnitude / Mathf.Max(1, priority);

        int leastImportantIndex = -1;
        float leastImportanceValue = float.MaxValue;

        // Find an available audio source to use
        for (int i = 0; i < _pool.Count; i++)
        {
            AudioPoolItem poolItem = _pool[i];

            // If we find an unused sound source, use it!
            if (!poolItem.Playing)
                return ConfigurePoolObject(i, mixerGroupName, clip, position, volume, pitch, spatialBlend, unimportance);
            else
            // We have a pool item that is less important than the one we are going to play
            if (poolItem.Unimportance > leastImportanceValue)
            {
                // Record the least important sound we have found so far
                // as a candidate to relace with our new sound request
                leastImportanceValue = poolItem.Unimportance;
                leastImportantIndex = i;
            }
        }

        // If we get here, all sounds are being used, but we know the least important sound currently being played
        // ... so if it is less important than OUR sound request, overwrite that sound source with our new request
        if (leastImportanceValue > unimportance)
            return ConfigurePoolObject(leastImportantIndex, mixerGroupName, clip, position, volume, pitch, spatialBlend, unimportance);


        // Could not be played (no sound in the pool available)
        return 0;
    }



    // -------------------------------------------------------------------------------
    // Name	:	PlayOneShotSound, after a delay
    // Desc	:	Queue up a one shot sound to be played after a number of seconds
    // -------------------------------------------------------------------------------
    public IEnumerator PlayOneShotSoundDelayed(string mixerGroupName, AudioClip clip, Vector3 position, float volume, float spatialBlend, float duration, int priority = 128)
    {
        yield return new WaitForSeconds(duration);
        PlayOneShotSound(mixerGroupName, clip, position, volume, spatialBlend, priority);
    }



    // -------------------------------------------------------------------------------
    // Name	:	PlayAudioCue    (NOTE: I'd say to use this one most of the time, nice and simple!!)
    // Desc	:	A nice short simple method to play an AudioCue, and just pass in the Cue and Position
    // -------------------------------------------------------------------------------
    public ulong PlayAudioCue(AudioCue audioCue, Vector3 position)
    {
        // Do nothing if track does not exist, clip is null or volume is zero
        if (!_groups.ContainsKey(audioCue.mixerGroup) || audioCue.audioClips == null  || audioCue.volume.Equals(0.0f)) return 0;

        // select a random audio clip from the audioCue
        AudioClip clip = audioCue.audioClips[Random.Range(0, audioCue.audioClips.Count)];

        float unimportance = (_listenerPos.position - position).sqrMagnitude / Mathf.Max(1, audioCue.priority);

        int leastImportantIndex = -1;
        float leastImportanceValue = float.MaxValue;

        // Find an available audio source to use
        for (int i = 0; i < _pool.Count; i++)
        {
            AudioPoolItem poolItem = _pool[i];
            
            // If we find an unused sound source, use it!
            if (!poolItem.Playing)
                return ConfigurePoolObject(i, audioCue.mixerGroup, clip, position, audioCue.volume, audioCue.pitch, audioCue.spatialBlend, unimportance);
            else
            // We have a pool item that is less important than the one we are going to play
            if (poolItem.Unimportance > leastImportanceValue)
            {
                // Record the least important sound we have found so far
                // as a candidate to relace with our new sound request
                leastImportanceValue = poolItem.Unimportance;
                leastImportantIndex = i;
            }
        }

        // If we get here, all sounds are being used, but we know the least important sound currently being played
        // ... so if it is less important than OUR sound request, overwrite that sound source with our new request
        if (leastImportanceValue > unimportance)
            return ConfigurePoolObject(leastImportantIndex, audioCue.mixerGroup, clip, position, audioCue.volume, audioCue.pitch, audioCue.spatialBlend, unimportance);


        // Could not be played (no sound in the pool available)
        return 0;
    }


}
