﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;
using UnityEngine.AI;
using UnityEngine.UI;
using System;

public class GameManager : MonoBehaviour
{
    // Statics
    private static GameManager _instance = null;
    public static GameManager Get
    {
        get
        {
            if (_instance == null)
                _instance = (GameManager)FindObjectOfType(typeof(GameManager));
            return _instance;
        }
    }

    // GameState enum
    public enum GameState { MainMenu, PauseMenu, InGame }
    public GameState gameState = GameState.MainMenu;

    // Menu references
    public GameObject menuRootBase;
    public GameObject menuRootMain;
    public GameObject menuRootOptions;
    public GameObject button_Play;
    public GameObject button_Resume;

    // Game scene references
    private GameObject gameSceneRootObject;

    // Lists of pausable elements
    List<Behaviour> m_pausedComponents = new List<Behaviour>();
    List<ParticleSystem> m_pausedParticleSystems = new List<ParticleSystem>();
    List<AudioListener> m_pausedAudioListeners = new List<AudioListener>();
    // rigid body 3D elements
    List<Rigidbody> m_paused3DBodies = new List<Rigidbody>();
    List<Vector3> m_paused3DBodyVelocities = new List<Vector3>();
    List<Vector3> m_paused3DBodyAngularVelocities = new List<Vector3>();
    // rigid body 2D elements
    List<Rigidbody2D> m_paused2DBodies = new List<Rigidbody2D>();
    List<Vector2> m_paused2DBodyVelocities = new List<Vector2>();
    List<float> m_paused2DBodyAngularVelocities = new List<float>();
    // nav agent elements
    List<NavMeshAgent> m_pausedAgents = new List<NavMeshAgent>();
    List<NavMeshPath> m_pausedAgentPaths = new List<NavMeshPath>();
    List<Vector3> m_pausedAgentVelocities = new List<Vector3>();
    List<Vector3> m_pausedAgentDestinations = new List<Vector3>();
    // sound sources
    List<AudioSource> m_pausedAudioSources = new List<AudioSource>();

    // Player Prefs
    private float audioVolMaster;
    private float audioVolMusic;
    private float audioVolSFX;
    private float audioVolVoice;
    private float controlsFOV;

    // Options Menus
    public Slider sliderVolMaster;
    public Slider sliderVolMusic;
    public Slider sliderVolSFX;
    public Slider sliderVolVoice;


    public void Start()
    {
        // Load in player preferences
        LoadPlayerPrefs();
        
        //enable the Options Menu, so we can initialise it  (we will auto-disable it again afterwards)
        menuRootOptions.SetActive(true);
        Invoke("InitialiseSettingsSliderValues", 0.02f);
        Invoke("InitialiseAudioMixerLevels", 0.03f);
        
        // start at the MainMenu
        GotoMainMenu();

        // HACK: Check if user is trying to directly play a game scene in the editor, if so set GameState.InGame and disable menus
        DisableMenuIfUserIsEditingGameScene();
    }



    private void LoadPlayerPrefs()
    {
        //audio
        audioVolMaster = PlayerPrefs.GetFloat("audioVolMaster", 0.8f);
        audioVolMusic = PlayerPrefs.GetFloat("audioVolMusic", 0.6f);
        audioVolSFX = PlayerPrefs.GetFloat("audioVolSFX", 0.8f);
        audioVolVoice = PlayerPrefs.GetFloat("audioVolVoice", 0.8f);
        //controls
        controlsFOV = PlayerPrefs.GetFloat("controlsFOV", 65f);

        // Example Code: Load pref and convert int into a bool (player pref doesn't natively support bool type)
        //useCrosshair = (PlayerPrefs.GetInt("useCrosshair", 1) == 1) ? true : false;

        // Example Code: Set a UI checkbox to the correct state based on pref
        //UIControl.toggle_crosshairControls.isOn = useCrosshair;        
    }


    private void InitialiseAudioMixerLevels()
    {
        // set Audio Mixer levels to PlayerPref values, immediately
        AudioManager.Get.SetTrackVolume("Master", Mathf.Log10(audioVolMaster) * 20, 0.0f);
        AudioManager.Get.SetTrackVolume("Music", Mathf.Log10(audioVolMusic) * 20, 0.0f);
        AudioManager.Get.SetTrackVolume("SFX", Mathf.Log10(audioVolSFX) * 20, 0.0f);
        AudioManager.Get.SetTrackVolume("Voice", Mathf.Log10(audioVolVoice) * 20, 0.0f);
    }


    private void InitialiseSettingsSliderValues()
    {
        // update audio settings sliders with player pref values
        sliderVolMaster.value = audioVolMaster;
        sliderVolMusic.value = audioVolMusic;
        sliderVolSFX.value = audioVolSFX;
        sliderVolVoice.value = audioVolVoice;
        
        //disable the Options Menu
        menuRootOptions.SetActive(false);
    }

    

    public void DisableMenuIfUserIsEditingGameScene()
    {
        // If "GameSceneRoot" object exists on startup, the user is editing a scene in the Editor, so go directly to GameState.InGame
        gameSceneRootObject = GameObject.Find("GameSceneRoot");
        if (gameSceneRootObject != null)
        {
            gameState = GameState.InGame;
            DisableMenus();
            Debug.Log("HACK: Disabled menus as a convenience because GameSceneRoot was detected when entering PlayMode.");
        }

        AudioManager.Get.FindGameAudioPoolParent();
    }



    public void Update()
    {
        CheckForPauseMenuCommand();
    }


    public void GotoOptionsMenu()
    {
        menuRootBase.SetActive(true);
        menuRootMain.SetActive(false);
        menuRootOptions.SetActive(true);
    }

    public void GotoMainMenu()
    {
        menuRootBase.SetActive(true);
        menuRootMain.SetActive(true);
        menuRootOptions.SetActive(false);
    }



    public void CheckForPauseMenuCommand()
    {
        if (Input.GetKeyDown(KeyCode.Escape))
        {
            if (gameState != GameState.MainMenu)
            {
                TogglePauseMenu();
            }
        }
    }

    

    public void PauseGame()
    {
        gameState = GameState.PauseMenu;
        EnableMenus();
        ShowResumeButton();

        // not really required, but does handle pausing and unpausing co-routines, which is nice
        //TimeScaleFreeze();

        gameSceneRootObject = GameObject.Find("GameSceneRoot");

        // Get a reference to all transforms that are a child of gameSceneRootObject (all objects in game scene)
        Transform[] allTransforms = gameSceneRootObject.GetComponentsInChildren<Transform>();
        // Make array of GameObjects from those transforms
        GameObject[] allObjects = new GameObject[allTransforms.Length];
        // Put the game objects into the array
        for (int i = 0; i < allTransforms.Length; i++)
        {
            allObjects[i] = allTransforms[i].gameObject;
        }

        // Declare temp objects
        GameObject obj = null;
        Behaviour[] tmpComponents = null;
        Behaviour tmpComponent = null;
        ParticleSystem[] tmpParticles = null;
        ParticleSystem tmpParticle = null;
        AudioListener[] tmpAudioListeners = null;
        AudioListener tmpAudioListener = null;
        AudioSource[] tmpAudioSources = null;
        AudioSource tmpAudioSource = null;
        int tmpComponentCount = 0;
        int count = allObjects.Length;


        // Go through all objects currently active in the hierarchy, and pause the relevant elements
        for (int i = 0; i < count; ++i)
        {
            obj = allObjects[i];
            if (obj.activeInHierarchy)
            {
                // TO DO:  Add back in support for RigidBody2D as well!

                // Pause 3D RigidBodies
                Rigidbody body = obj.GetComponent<Rigidbody>();
                if (body != null && body.isKinematic == false)
                {
                    m_paused3DBodyVelocities.Add(body.velocity);
                    m_paused3DBodyAngularVelocities.Add(body.angularVelocity);
                    m_paused3DBodies.Add(body);
                    body.isKinematic = true;
                    body.velocity = Vector3.zero;
                    body.angularVelocity = Vector3.zero;
                }

                // Pause 2D RigidBodies
                Rigidbody2D body2D = obj.GetComponent<Rigidbody2D>();
                if (body2D != null && body2D.isKinematic == false)
                {
                    m_paused2DBodyVelocities.Add(body2D.velocity);
                    m_paused2DBodyAngularVelocities.Add(body2D.angularVelocity);
                    m_paused2DBodies.Add(body2D);
                    body2D.isKinematic = true;
                    body2D.velocity = Vector2.zero;
                    body2D.angularVelocity = 0;
                }

                // Pause "Behaviours" (custom Component Scripts)
                tmpComponents = obj.GetComponents<Behaviour>();
                tmpComponentCount = tmpComponents.Length;

                for (int j = 0; j < tmpComponentCount; ++j)
                {
                    tmpComponent = tmpComponents[j];
                    if (tmpComponent.enabled && (tmpComponent is MonoBehaviour || tmpComponent is Animator)) // && IsBehaviourPausable
                    {
                        m_pausedComponents.Add(tmpComponent);
                        tmpComponent.enabled = false;
                    }
                }

                // Pause Particles
                tmpParticles = obj.GetComponents<ParticleSystem>();
                tmpComponentCount = tmpParticles.Length;
                for (int j = 0; j < tmpComponentCount; ++j)
                {
                    tmpParticle = tmpParticles[j];
                    if (tmpParticle.isPlaying && tmpParticle.isPaused == false)
                    {
                        tmpParticle.Pause();
                        m_pausedParticleSystems.Add(tmpParticle);
                    }
                }

                // Pause Audio Sources
                tmpAudioSources = obj.GetComponents<AudioSource>();
                tmpComponentCount = tmpAudioSources.Length;

                for (int j = 0; j < tmpComponentCount; ++j)
                {
                    tmpAudioSource = tmpAudioSources[j];
                    if (tmpAudioSource.isPlaying)
                    {
                        m_pausedAudioSources.Add(tmpAudioSource);
                        tmpAudioSource.Pause();
                    }
                }

                // Pause Audio Listeners  (disable game camera Audio Listener)
                tmpAudioListeners = obj.GetComponents<AudioListener>();
                tmpComponentCount = tmpAudioListeners.Length;
                for (int j = 0; j < tmpComponentCount; ++j)
                {
                    tmpAudioListener = tmpAudioListeners[j];
                    if (tmpAudioListener.enabled == true)
                    {
                        tmpAudioListener.enabled = false;
                        m_pausedAudioListeners.Add(tmpAudioListener);
                    }
                }

                // Pause Nav Agents
                NavMeshAgent agent = obj.GetComponent<NavMeshAgent>();
                if (agent != null && agent.enabled == true)
                {
                    // TO DO:  Clean this up once I'm sure Nav Agents Pause & Unpause perfectly.
                    // TO DO:  Including cleaning out any uneeded variables (arralys, lists) at the top!
                    /*  
                    // store agent values
                    m_pausedAgents.Add(agent);
                    m_pausedAgentVelocities.Add(agent.velocity);
                    m_pausedAgentPaths.Add(agent.path);
                    m_pausedAgentDestinations.Add(agent.destination);
                    // pause agent values
                    agent.velocity = Vector3.zero;
                    agent.ResetPath();
                    agent.enabled = false;
                    */

                    m_pausedAgents.Add(agent);
                    m_pausedAgentVelocities.Add(agent.velocity);
                    //agent.velocity = Vector3.zero;
                    //agent.ResetPath();
                    // TO DO:  Is it better to just DISABLE the agent, or set velocity & path to "0" instead?
                    agent.enabled = false;
                }
                
            }
        }
    }

    

    public void UnPauseGame()
    {
        // TO DO:  Should use COUNTER system to know when to unpause (in future)
        //if (m_pauseSources.Count == 0) {  

        // Restore the Behaviours
        Behaviour behaviour = null;
        int count = m_pausedComponents.Count;
        for (int i = 0; i < count; ++i)
        {
            behaviour = m_pausedComponents[i];
            if (behaviour != null)
                behaviour.enabled = true;
        }
        m_pausedComponents.Clear();

        // Restore the 3D RigidBodies
        Rigidbody body = null;
        count = m_paused3DBodies.Count;
        for (int i = 0; i < count; ++i)
        {
            body = m_paused3DBodies[i];
            if (body != null)
            {
                body.isKinematic = false;
                body.velocity = m_paused3DBodyVelocities[i];
                body.angularVelocity = m_paused3DBodyAngularVelocities[i];
                body.WakeUp();
            }
        }
        m_paused3DBodies.Clear();
        m_paused3DBodyVelocities.Clear();
        m_paused3DBodyAngularVelocities.Clear();

        // Restore 2D RigidBodies
        Rigidbody2D body2D = null;
        count = m_paused2DBodies.Count;
        for (int i = 0; i < count; ++i)
        {
            body2D = m_paused2DBodies[i];
            if (body2D != null)
            {
                body2D.isKinematic = false;
                body2D.velocity = m_paused2DBodyVelocities[i];
                body2D.angularVelocity = m_paused2DBodyAngularVelocities[i];
                body2D.WakeUp();
            }
        }
        m_paused2DBodies.Clear();
        m_paused2DBodyVelocities.Clear();
        m_paused2DBodyAngularVelocities.Clear();


        // Restore the Particle Systems
        count = m_pausedParticleSystems.Count;
        for (int i = 0; i < count; ++i)
        {
            if (m_pausedParticleSystems[i] != null)
                m_pausedParticleSystems[i].Play();
        }
        m_pausedParticleSystems.Clear();

        // Restore Audio Sources
        AudioSource audioSource = null;
        count = m_pausedAudioSources.Count;
        for (int i = 0; i < count; ++i)
        {
            audioSource = m_pausedAudioSources[i];
            if (audioSource != null)
                audioSource.UnPause();
        }
        m_pausedAudioSources.Clear();

        // Restore Audio Listeners
        AudioListener audioListener = null;
        count = m_pausedAudioListeners.Count;
        for (int i = 0; i < count; ++i)
        {
            audioListener = m_pausedAudioListeners[i];
            if (audioListener != null)
                audioListener.enabled = true;
        }
        m_pausedAudioListeners.Clear();

        // Restore Nav Agents
        NavMeshAgent agent = null;
        count = m_pausedAgents.Count;
        for (int i = 0; i < count; ++i)
        {
            agent = m_pausedAgents[i];
            if (agent != null)
            {

                //agent.isStopped = false;
                agent.enabled = true;
                agent.velocity = m_pausedAgentVelocities[i];


                // TO DO:  MUZ:  Tidy this up properly after I've tested the system with fully animated skinned mesh agents!

                /*
                agent.velocity = m_pausedAgentVelocities[i];
                agent.enabled = true;
                agent.SetPath(m_pausedAgentPaths[i]);
                */


                /*
                // restore agent values
                if (agent.destination == m_pausedAgentDestinations[i])
                {
                    agent.SetPath(m_pausedAgentPaths[i]);
                }
                else
                {
                    // fallback in case the agent has somehow "lost" its destination
                    agent.SetDestination(m_pausedAgentDestinations[i]);
                }
                agent.velocity = m_pausedAgentVelocities[i];
                */

            }
        }
        // clear out all "pauised lists" for agents
        m_pausedAgents.Clear();
        m_pausedAgentVelocities.Clear();
        m_pausedAgentDestinations.Clear();
        m_pausedAgentPaths.Clear();

        // FINAL STEPS

        // Disable Menu objects at the root
        menuRootBase.SetActive(false);

        // Restore normal time scale  (not really required, but does handle pausing and unpausing co-routines, which is nice)
        //TimeScaleUnfreeze();

        gameState = GameState.InGame;
    }


    public void TogglePauseMenu()
    {
        if (gameState == GameState.InGame)
        { PauseGame(); }
        else
        { UnPauseGame(); }
    }


    public void ShowResumeButton()
    {
        button_Play.SetActive(false);
        button_Resume.SetActive(true);
    }


    public void ChangeScene(string sceneName)
    {
        SceneManager.LoadScene(sceneName, LoadSceneMode.Additive);
        gameState = GameState.InGame;
    }


    public void EnableMenus()
    {
        menuRootBase.SetActive(true);
    }


    public void DisableMenus()
    {
        menuRootBase.SetActive(false);
    }


    public void QuitGame()
    {
        Application.Quit();
    }


    public void TimeScaleFreeze()
    {
        Time.timeScale = 0f;
    }


    public void TimeScaleUnfreeze()
    {
        Time.timeScale = 1f;
    }


}
